from pathlib import Path

# package
ROOT_PATH = Path(__file__).parent.parent
PACKAGE_NAME = "snips_nlu"
DATA_PACKAGE_NAME = "data"
DATA_PATH = ROOT_PATH / PACKAGE_NAME / DATA_PACKAGE_NAME
PACKAGE_PATH = ROOT_PATH / PACKAGE_NAME

# result
RES_INPUT = "input"
RES_INTENT = "intent"
RES_SLOTS = "slots"
RES_INTENT_NAME = "intentName"
RES_PROBA = "probability"
RES_SLOT_NAME = "slotName"
RES_ENTITY = "entity"
RES_VALUE = "value"
RES_RAW_VALUE = "rawValue"
RES_MATCH_RANGE = "range"

# miscellaneous
AUTOMATICALLY_EXTENSIBLE = "automatically_extensible"
USE_SYNONYMS = "use_synonyms"
SYNONYMS = "synonyms"
DATA = "data"
INTENTS = "intents"
ENTITIES = "entities"
ENTITY = "entity"
ENTITY_KIND = "entity_kind"
RESOLVED_VALUE = "resolved_value"
SLOT_NAME = "slot_name"
TEXT = "text"
UTTERANCES = "utterances"
LANGUAGE = "language"
VALUE = "value"
NGRAM = "ngram"
TOKEN_INDEXES = "token_indexes"
CAPITALIZE = "capitalize"
UNKNOWNWORD = "unknownword"
VALIDATED = "validated"
START = "start"
END = "end"
BUILTIN_ENTITY_PARSER = "builtin_entity_parser"
CUSTOM_ENTITY_PARSER = "custom_entity_parser"
MATCHING_STRICTNESS = "matching_strictness"

# resources
RESOURCES = "resources"
METADATA = "metadata"
STOP_WORDS = "stop_words"
NOISE = "noise"
GAZETTEERS = "gazetteers"
STEMS = "stems"
CUSTOM_ENTITY_PARSER_USAGE = "custom_entity_parser_usage"
WORD_CLUSTERS = "word_clusters"
GAZETTEER_ENTITIES = "gazetteer_entities"

# builtin entities
SNIPS_AMOUNT_OF_MONEY = "snips/amountOfMoney"
SNIPS_DATETIME = "snips/datetime"
SNIPS_DURATION = "snips/duration"
SNIPS_NUMBER = "snips/number"
SNIPS_ORDINAL = "snips/ordinal"
SNIPS_PERCENTAGE = "snips/percentage"
SNIPS_TEMPERATURE = "snips/temperature"

# languages
LANGUAGE_DE = "de"
LANGUAGE_EN = "en"
LANGUAGE_ES = "es"
LANGUAGE_FR = "fr"
LANGUAGE_IT = "it"
LANGUAGE_JA = "ja"
LANGUAGE_KO = "ko"
LANGUAGE_PT_BR = "pt_br"
LANGUAGE_PT_PT = "pt_pt"



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -msnips_app_helpers` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``snips_app_helpers.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``snips_app_helpers.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import pathlib

import click

from . import specs


@click.group()
@click.option('--debug/--no-debug', default=False)
def main(debug):
    if debug:
        click.echo('Debug mode is %s' % ('on' if debug else 'off'))


@main.group()
def spec():
    pass


@spec.command()
@click.option('-aj',
              '--assistant_json',
              default=pathlib.Path('/usr/share/snips/assistant/'),
              type=pathlib.Path)
@click.option('-ad',
              '--app_dir',
              default=pathlib.Path('/var/lib/snips/skills'),
              type=pathlib.Path)
def check(assistant_json, app_dir):
    if not assistant_json.exists():
        click.echo(click.style('"%s" does not seems to be an existing file'
                               % str(assistant_json), fg='red'))
        return

    if not app_dir.exists():
        click.echo(click.style('"%s" does not seems to be an existing folder'
                               % str(app_dir), fg='red'))
        return
    click.echo(('Analysing spec for:\n'
                '\tassistant: %s\n'
                '\tapp dir: %s') % (
                    click.style(str(assistant_json), fg="cyan"),
                    click.style(str(app_dir), fg="cyan")
                ))
    report_messages = specs.AssistantSpec.load(assistant_json).check(app_dir)
    SpecReportCli(report_messages).show()


class SpecReportCli(specs.message.Report):

    def _print_list(self, msg_type, messages, color):
        click.echo(click.style(str(
            "\n" +
            msg_type.print_list(
                messages
            )
        ), fg=color))

    def show(self):
        # infos on top
        for msg_type, messages in self.grouped_messages.items():
            if issubclass(msg_type, specs.message.Info):
                self._print_list(msg_type, messages, 'cyan')

        for msg_type, messages in self.grouped_messages.items():
            if issubclass(msg_type, specs.message.Warning):
                self._print_list(msg_type, messages, 'yellow')
            elif issubclass(msg_type, specs.message.Error):
                self._print_list(msg_type, messages, 'red')



# coding=utf-8

from collections import defaultdict

import yaml

from .. import utils
from . import message


class Quantifier(utils.BaseObj):

    def __init__(self, quantifier_str):
        self._quantifier = str(quantifier_str)

    def contains(self, quantity):
        if self._quantifier == '+':
            if quantity <= 0:
                return False
        elif self._quantifier == '*':
            if quantity < 0:
                return False
        elif self._quantifier == '?':
            if quantity < 0 or quantity > 1:
                return False
        elif self._quantifier.startswith('{') and self._quantifier.endswith('}'):
            # implement {1,10}
            start, end = map(int,
                             self._quantifier.replace('{', '').replace('}', '').split(','))
            if not (start <= quantity <= end):
                return False
        else:
            try:
                qte = int(self._quantifier)
                if quantity != qte:
                    return False
            except ValueError:
                return False
        return True

    def __str__(self):
        return "<%s %s>" % (
            self.__class__.__name__,
            self._quantifier,
        )


class CoverageRuleBlock(utils.BaseObj):

    def __init__(self, slot_name, quantifier):
        self.slot_name = slot_name
        self.quantifier = quantifier

    def cover(self, slot_name, slot_qte):
        return (
            slot_name == self.slot_name and self.quantifier.contains(slot_qte)
        )

    @classmethod
    def load_from_spec(cls, slot_name, quantifier_str):
        return cls(
            slot_name,
            quantifier=Quantifier(quantifier_str),
        )

    def __str__(self):
        return "<%s slot_name=%s quantifier=%s>" % (
            self.__class__.__name__,
            self.slot_name,
            self.quantifier,
        )


class CoverageRule(utils.BaseObj):

    def __init__(self, rule_blocks):
        self.rule_blocks = rule_blocks

    @classmethod
    def load_from_spec(cls, rule_blocks):
        return cls(
            list(CoverageRuleBlock.load_from_spec(slot_name, quantifier)
            for slot_name, quantifier in rule_blocks)
        )

    def cover(self, slot_occurences):
        if not slot_occurences:
            True

        # TODO fix: check that every pattern as been consumed
        # for all elm of sequence
        return all(
            # at least one coverage block match the elm
            any(
                rule_block.cover(slot_name, slot_qte)
                for rule_block in self.rule_blocks
            )
            for slot_name, slot_qte in slot_occurences
        )

    def __str__(self):
        return "<%s rules=%s>" % (
            self.__class__.__name__,
            self.rule_blocks
        )


class CoverageIntentSpec(utils.BaseObj):

    def __init__(self, intent_name, rules):
        self.name = intent_name
        self.rules = rules

    @classmethod
    def load_from_spec(cls, intent_name, intent_cov_spec):
        return cls(intent_name, [
            CoverageRule.load_from_spec(rule_sequence)
            for rule_sequence in intent_cov_spec
        ])

    def covered_slots_sequence(self, slots_sequences):
        # iter over coverage rules and check it is a valid
        covered_slot_seq = set()
        for slots_sequence in slots_sequences:
            for rule in self.rules:
                if rule.cover(slots_sequence):
                    covered_slot_seq.add(slots_sequence)
                    break
        return covered_slot_seq

    def uncovered_slots_sequence(self, slots_sequences):
        return set(tuple(seq_case) for seq_case in slots_sequences).difference(
            self.covered_slots_sequence(slots_sequences))

    def check(self, dataset_intent, action_spec_filepath):
        report_msgs = set()
        # Check full coverage
        uncovered_seq = self.uncovered_slots_sequence(dataset_intent.slots_sequences)
        for slot_seq in uncovered_seq:
            report_msgs.add(message.CoverageSlotSeq(
                spec_filepath=action_spec_filepath,
                intent_name=self.name,
                slots_sequences='[%s]' % ', '.join(
                    "%dx %s" % (qte, sn) for sn, qte in slot_seq
                )
            ))
        return report_msgs

    def __str__(self):
        return "<%s rules=%s>" % (
            self.__class__.__name__,
            self.rules
        )


class CoverageSpec(utils.BaseObj):

    def __init__(self, intents_specs, action_spec=None):
        self.intents = intents_specs
        self._action_spec = action_spec

    def set_linked_action_spec(self, action_spec):
        self._action_spec = action_spec

    @property
    def action_spec(self):
        if self._action_spec is None:
            raise ValueError(
                "call to action_spec before setting it with "
                "set_linked_action_spec or at init time."
            )
        return self._action_spec

    @classmethod
    def load_from_spec(cls, spec_coverage_dic, action_spec=None):
        return CoverageSpec(
            {
                intent_name: CoverageIntentSpec.load_from_spec(
                    intent_name,
                    intent_cov_spec
                ) if intent_cov_spec else None
                for intent_name, intent_cov_spec in spec_coverage_dic.items()
            },
            action_spec=action_spec
        )

    def check_slot_consistancy(self, assistant_intent):
        report_msgs = set()

        if not assistant_intent.slots:
            assistant_intent_slot_names, assistant_intent_slot_type = [], []
        else:
            assistant_intent_slot_names, assistant_intent_slot_type = zip(*[
                (slot.get('name'), slot.get('entityId'))
                for slot in assistant_intent.slots
            ])

        for intent_name, intent_slot_config in self.intents.items():
            if intent_name not in assistant_intent.name:
                continue
            if not intent_slot_config:
                continue
            for rule in intent_slot_config.rules:
                for rule_block in rule.rule_blocks:
                    slot_name = rule_block.slot_name
                    try:
                        idx = list(assistant_intent_slot_names).index(slot_name)
                        if self.action_spec.slots:
                            slot_type = self.action_spec.slots.get(slot_name)
                            expected_slot_type = list(assistant_intent_slot_type)[idx]
                            if slot_type and not expected_slot_type.startswith(slot_type):
                                report_msgs.add(message.InvalidSlotType(
                                    spec_filepath=self.action_spec.rel_filepath,
                                    intent_name=intent_name,
                                    slot_name=slot_name,
                                    slot_type=slot_type,
                                    expected_slot_type=expected_slot_type,
                                ))
                    except ValueError:
                        report_msgs.add(message.MissingSlot(
                            spec_filepath=self.action_spec.rel_filepath,
                            intent_name=intent_name,
                            slot_name=slot_name,
                            assistant_slot_names=assistant_intent_slot_names,
                        ))
        return report_msgs

    def check(self, assistant):
        # intents_coverage, dataset
        report_msgs = set()
        intents_coverage = defaultdict(list)
        for action_intent_name, slot_spec in self.intents.items():
            if action_intent_name in assistant.intents:
                if slot_spec:
                    dataset_intent = assistant.dataset.intent_per_name.get(
                        action_intent_name
                    )
                    if dataset_intent:
                        intent_messages = set()
                        intent_messages.update(slot_spec.check(
                            dataset_intent=dataset_intent,
                            action_spec_filepath=self.action_spec.rel_filepath
                        ))
                        intent_messages.update(
                            self.check_slot_consistancy(
                                assistant.intents[action_intent_name]
                            )
                        )
                        if not intent_messages:
                            intent_messages.add(
                                message.CorrectlyLinked(
                                    spec_filepath=self.action_spec.rel_filepath,
                                    action_dir=self.action_spec.action_dir,
                                    intent_name=action_intent_name
                                )
                            )
                        report_msgs.update(intent_messages)
                    else:
                        # incoherence between dataset.json and assistant.json !
                        report_msgs.add(message.IncoherentAssistantDatasetIntent(
                            spec_filepath=self.action_spec.rel_filepath,

                        ))
                else:
                    report_msgs.add(
                        message.GenericCoveredIntent(
                            spec_filepath=self.action_spec.rel_filepath,
                            action_dir=self.action_spec.action_dir.name,
                            intent_name=action_intent_name
                        )
                    )
                intents_coverage[action_intent_name].append(self.action_spec.name)
            else:
                report_msgs.add(message.IntentNotInAssistant(
                    intent_name=action_intent_name,
                    action_name=self.action_spec.name,
                ))

        return (report_msgs, intents_coverage)

    def __str__(self):
        return "<%s cover_rules=%s>" % (
            self.__class__.__name__,
            self.intents,
        )


class ActionSpec(utils.BaseObj):

    FILENAME = "spec.yml"
    SUFFIX_EXTERAL_FILENAME = "." + FILENAME

    ATTRS = [
        "name",
        "supported_snips_versions",
        "version",
        "coverage",
        "_coverage_dic",
        "slots",
        "udpated_at"
    ]

    def __init__(self,
                 filepath,
                 action_dir,
                 have_spec=False, **kwargs):
        self.filepath = filepath
        self.action_dir = action_dir
        self.have_spec = have_spec
        for k, v in kwargs.items():
            if k in self.ATTRS:
                setattr(self, k, v)

    @property
    def rel_filepath(self):
        return self.filepath.relative_to(self.action_dir.parent)

    @classmethod
    def load(cls, filepath):
        if not filepath.is_file():
            return ActionSpec(
                filepath,
                filepath.parent,
                have_spec=False
            )
        with filepath.open('r') as fh:
            action_spec_dic = yaml.load(fh, Loader=yaml.FullLoader)

        if not action_spec_dic:
            raise ValueError("Spec %s is Empty !" % filepath)

        action_dir = filepath.parent
        if str(filepath).endswith(cls.SUFFIX_EXTERAL_FILENAME):
            candidate_action_dir = action_dir.parent / filepath.stem.replace('.spec', '')
            if candidate_action_dir.is_dir():
                action_dir = candidate_action_dir
            else:
                raise ValueError(
                    "A spec is defined that does not exists in any action"
                    "directory: not found dir %s" % candidate_action_dir
                )

        kwargs = {
            attr: action_spec_dic.get(attr)
            for attr in cls.ATTRS
        }
        kwargs["_coverage_dic"] = kwargs["coverage"]
        kwargs["coverage"] = CoverageSpec.load_from_spec(kwargs["coverage"])

        action_spec = ActionSpec(
            filepath=filepath,
            action_dir=action_dir,
            have_spec=True,
            **kwargs
        )
        kwargs["coverage"].set_linked_action_spec(action_spec)
        return action_spec

    @classmethod
    def load_all_in_action_code_dir(cls, actions_dir):
        actions_specs = []
        for action_dir in actions_dir.iterdir():
            if action_dir.is_dir():
                base_filepath = action_dir / cls.FILENAME
                actions_specs.append(
                    ActionSpec.load(base_filepath)
                )
                for another_spec in action_dir.glob('*.' + cls.FILENAME):
                    actions_specs.append(
                        ActionSpec.load(another_spec)
                    )
        return actions_specs

    def check(self, assistant):
        report_msgs = set()
        report_msgs.add(
            message.DetectedSpec(
                spec_filepath=self.rel_filepath,
                action_dir=self.action_dir.name,
            )
        )
        if not self.have_spec:
            report_msgs.add(message.NoSpec(
                action_dir=self.action_dir.name
            ))
            return (report_msgs, {})

        cov_msgs, intents_coverage = self.coverage.check(assistant)
        report_msgs.update(cov_msgs)
        return (report_msgs, intents_coverage)

    def __str__(self):
        return "<%s action_dir='%s' %s>" % (
            self.__class__.__name__,
            self.action_dir,
            ("no-spec" if not self.have_spec else " ".join(
                "%s=%s" % (attr, getattr(self, attr)) for attr in self.ATTRS)
            )
        )



import os
from click.testing import CliRunner
import pytest
from pathlib import Path

from snips_app_helpers.cli import main
from snips_app_helpers.specs import action
from snips_app_helpers.specs import assistant
from snips_app_helpers.specs import message
from snips_app_helpers.snips.dataset import Dataset


filedir = os.path.dirname(os.path.abspath(__file__))


def coverage_check(action_spec_path):
    ass = assistant.Assistant.load(
        Path(os.path.join(filedir, "./fixtures/assistant_1/assistant.json"))
    )
    action_spec = action.ActionSpec.load(Path(
        os.path.join(filedir, "./fixtures/%s" % action_spec_path)
    ))
    return action.CoverageSpec.load_from_spec(
        action_spec._coverage_dic,
        action_spec
    ).check(ass)


def assistant_check(actions_dirname):
    return assistant.AssistantSpec.load(
        Path(os.path.join(filedir, "./fixtures/assistant_1/assistant.json"))
    ).check(
        Path(os.path.join(filedir, "./fixtures/%s" % actions_dirname))
    )


@pytest.mark.parametrize(
    "quantifier,number,is_in",
    [
        ("+", 1, True),
        ("+", 0, False),
        ("+", -1, False),
        ("?", 0, True),
        ("?", 1, True),
        ("?", 100, False),
        ("*", 0, True),
        ("*", 1, True),
        ("*", 100, True),
        ("*", -1, False),
        ("{1,3}", 100, False),
        ("{1,3}", 1, True),
        ("{1,3}", 0, False),
        ("{1,3}", 3, True),
        ("1", 100, False),
        ("1", 1, True),
        ("5", 5, True),
        ("a", 1, False),
    ]
)
def test_quantifier_check(quantifier, number, is_in):
    quanti = action.Quantifier(quantifier)
    assert quanti.contains(number) == is_in
    assert quantifier in str(quanti)


@pytest.mark.parametrize(
    "slot_name,quantifier,c_slot_name,c_slot_qte,is_in",
    [
        ("a", "?", "a", 1, True),
        ("a", "?", "b", 1, False),
        ("a", "?", "b", 0, False),
        ("ca", "+", "ca", 1, True),
        ("ca", "+", "ca", 0, False),
    ]
)
def test_coverage_block(slot_name, quantifier, c_slot_name, c_slot_qte, is_in):
    assert action.CoverageRuleBlock.load_from_spec(
        slot_name,
        quantifier
    ).cover(
        c_slot_name,
        c_slot_qte
    ) == is_in


@pytest.mark.parametrize(
    "rules_blocks,slot_occurences,is_in",
    [
        ([("a", "+"), ("b", "?")], [("a", 1), ("b", 1)], True),
        ([("a", "+"), ("b", "?")], [("a", 1), ("b", 2)], False),
        ([("a", "+"), ("b", "?")], [("a", 1)], True),
    ]
)
def test_coverage_rule(rules_blocks, slot_occurences, is_in):
    assert action.CoverageRule.load_from_spec(
        rules_blocks
    ).cover(slot_occurences) == is_in


def test_exec_coverage_intent_spec():
    dataset = Dataset.from_json(
        "en", Path(os.path.join(filedir, "./fixtures/assistant_1/dataset.json"))
    )
    ds_intent = dataset.intents[0]
    assert isinstance(action.CoverageIntentSpec.load_from_spec(
        ds_intent.name, [
            [("a", "+"), ("b", "?")]
        ]
    ).check(ds_intent, action_spec_filepath="bloadsa"), set)


def test_exec_coverage_spec():
    report_msgs, intents_coverage = coverage_check(
        action_spec_path="actions_1/Likhitha.Today/spec.yml"
    )
    assert isinstance(report_msgs, set)
    assert isinstance(intents_coverage, dict)


def test_check_detected_spec():
    assistant_spec = assistant.Assistant.load(
        Path(os.path.join(filedir, "./fixtures/assistant_1/assistant.json"))
    )

    fp = Path(os.path.join(filedir, "./fixtures/actions_1/Likhitha.Today/spec.yml"))
    report_msgs, intents_coverage = action.ActionSpec.load(fp).check(assistant_spec)

    assert isinstance(report_msgs, set)
    assert isinstance(intents_coverage, dict)
    detected_spec = next(
        report_msg
        for report_msg in report_msgs
        if isinstance(report_msg, message.DetectedSpec)
    )
    assert detected_spec.spec_filepath == fp.relative_to(
        fp.parent.parent
    )

def test_check_no_spec():
    assistant_spec = assistant.Assistant.load(
        Path(os.path.join(filedir, "./fixtures/assistant_1/assistant.json"))
    )

    action_dirname = "Likhitha.Today"
    fp = Path(os.path.join(filedir, "./fixtures/actions_1/%s/fake_spec.yml" % action_dirname))
    report_msgs, intents_coverage = action.ActionSpec.load(fp).check(assistant_spec)

    assert isinstance(report_msgs, set)
    assert isinstance(intents_coverage, dict)
    no_spec = next(
        report_msg
        for report_msg in report_msgs
        if isinstance(report_msg, message.NoSpec)
    )
    assert no_spec.action_dir == action_dirname


def test_check_intent_nin_assistant():
    report_msgs, intents_coverage = coverage_check(
        action_spec_path="actions_1/Marshall.Music_Player/spec.yml"
    )
    assert any(
        (
            isinstance(report_msg, message.IntentNotInAssistant) and
            report_msg.intent_name == 'MySuperFakeIntent'
        )
        for report_msg in report_msgs
    )


def test_check_not_covered_intent():
    report_msgs = assistant_check("actions_1")
    assert all(
        any(
            isinstance(report_msg, message.NotCoveredIntent) and report_msg.intent_name == intent_name
            for report_msg in report_msgs
        )
        for intent_name in [
            'getRandomTime',
            'bye', 'shiftDown', 'turnOff', 'temperatureConverter', 'getInfos', 'day', 'timeNow', 'date', 'setColor', 'currencyConverter', 'hello', 'lengthConverter', 'speedConverter', 'setBrightness', 'turnOn', 'areaFunction', 'getRandomDate', 'previousSong', 'setScene', 'weightConverter', 'speakerInterrupt']

    )


def test_check_coherent_assistant_dataset_intent():
    assert not any(isinstance(
        report_msg,
        message.IncoherentAssistantDatasetIntent
    ) for report_msg in assistant_check("actions_1"))


def test_check_missing_slot():
    assert any((
        isinstance(
            report_msg,
            message.MissingSlot
        ) and report_msg.slot_name == "numberx"
    ) for report_msg in assistant_check("actions_1"))


def test_intent_hooked_multiple_times():
    assert any((
        isinstance(
            report_msg,
            message.IntentHookedMultipleTimes
        ) and report_msg.intent_name == "getCurrentTime"
        and set(report_msg.action_names) == set(['Today', 'Calculation', 'Music Player'])
    ) for report_msg in assistant_check("actions_1"))


def test_check_coverage_slot_seq():
    assert any((
        isinstance(
            report_msg,
            message.CoverageSlotSeq
        ) and report_msg.intent_name == "volumeConverter"
    ) for report_msg in assistant_check("actions_1"))


def test_check_invalid_slot_type():
    assert any((
        isinstance(
            report_msg,
            message.InvalidSlotType
        ) and report_msg.slot_name == "percent"
    ) for report_msg in assistant_check("actions_1"))


def test_check_correctly_linked():
    assert any((
        isinstance(
            report_msg,
            message.CorrectlyLinked
        ) and report_msg.intent_name == "playArtist"
    ) for report_msg in assistant_check("actions_1"))


def test_check_generic_coverage():
    assert any((
        isinstance(
            report_msg,
            message.GenericCoveredIntent
        ) and report_msg.intent_name == 'playPlaylist'
    ) for report_msg in assistant_check("actions_1"))


# Tests END to END

def test_main():
    runner = CliRunner()
    result = runner.invoke(main, [])
    assert 'spec\n' in result.output
    assert result.exit_code == 0


def test_main_with_param():
    runner = CliRunner()
    result = runner.invoke(main, [
        "--debug",
        "spec",
        "check",
        "-aj",
        os.path.join(filedir, "./fixtures/assistant_1/assistant.json"),
        "--app_dir",
        os.path.join(filedir, "./fixtures/actions_1/")
    ])
    assert 'Detected spec:' in result.output
    assert result.exit_code == 0

    result = runner.invoke(main, [
        "spec",
        "check",
        "-aj",
        "./false_path",
        "--app_dir",
        os.path.join(filedir, "./fixtures/actions_1/")
    ])
    assert 'does not seems to be an existing file' in result.output
    assert result.exit_code == 0

    result = runner.invoke(main, [
        "spec",
        "check",
        "-aj",
        os.path.join(filedir, "./fixtures/assistant_1/assistant.json"),
        "--app_dir",
        "falsepath"
    ])
    assert 'does not seems to be an existing folder' in result.output
    assert result.exit_code == 0



# coding=utf-8

import json
import datetime
from collections import defaultdict

from . import ActionSpec
from . import message
from ..snips import Dataset
from ..snips import Assistant
from .. import utils


class AssistantSpec(utils.BaseObj):

    """ AssistantSpec

    Is here to allow to check all assistant specification against action codes
    """

    def __init__(self, assistant):
        self.assistant = assistant

    @classmethod
    def load(cls, filepath):
        return cls(Assistant.load(filepath))

    def compare_to_action_specs(self, action_spec_list):
        report_msgs = set()
        intents_coverage = defaultdict(list)
        for action_spec in action_spec_list:
            action_report_msgs, action_intents_coverage = action_spec.check(
                assistant=self.assistant
            )
            for cov_intent_name in action_intents_coverage:
                intents_coverage[cov_intent_name].append(action_spec.name)
            report_msgs.update(action_report_msgs)

        for intent_name, action_names in intents_coverage.items():
            if len(action_names) > 1:
                report_msgs.add(
                    message.IntentHookedMultipleTimes(
                        intent_name=intent_name, action_names=action_names
                    )
                )
        for not_covered_intent in set(self.assistant.intents).difference(intents_coverage):
            report_msgs.add(message.NotCoveredIntent(intent_name=not_covered_intent))
        return report_msgs

    def check(self, actions_dir):
        return self.compare_to_action_specs(
            ActionSpec.load_all_in_action_code_dir(actions_dir)
        )



from abc import ABCMeta, abstractmethod
from builtins import object
from io import IOBase
from collections import Counter

import yaml

from .constants import DATA, ENTITY, SLOT_NAME, TEXT, UTTERANCES
from .exceptions import IntentFormatError

from ... import utils


class Intent(utils.BaseObj):
    """Intent data of a :class:`.Dataset`

    Attributes:
        intent_name (str): name of the intent
        utterances (list of :class:`.IntentUtterance`): annotated intent
            utterances
        slot_mapping (dict): mapping between slot names and entities
    """

    def __init__(self, intent_name, utterances, slot_mapping=None, version=None,
                 id=None):
        if slot_mapping is None:
            slot_mapping = dict()
        self.name = intent_name
        self.utterances = utterances
        self.slot_mapping = slot_mapping
        self.version = version
        self.id = id
        self._complete_slot_name_mapping()
        self._ensure_entity_names()

    @classmethod
    def from_yaml(cls, yaml_dict):
        """Build an :class:`.Intent` from its YAML definition object

        Args:
            yaml_dict (dict or :class:`.IOBase`): object containing the YAML
                definition of the intent. It can be either a stream, or the
                corresponding python dict.

        Examples:
            An intent can be defined with a YAML document following the schema
            illustrated in the example below:

            >>> import io
            >>> from snips_nlu.common.utils import json_string
            >>> intent_yaml = io.StringIO('''
            ... # searchFlight Intent
            ... ---
            ... type: intent
            ... name: searchFlight
            ... slots:
            ...   - name: origin
            ...     entity: city
            ...   - name: destination
            ...     entity: city
            ...   - name: date
            ...     entity: snips/datetime
            ... utterances:
            ...   - find me a flight from [origin](Oslo) to [destination](Lima)
            ...   - I need a flight leaving to [destination](Berlin)''')
            >>> intent = Intent.from_yaml(intent_yaml)
            >>> print(json_string(intent.json, indent=4, sort_keys=True))
            {
                "utterances": [
                    {
                        "data": [
                            {
                                "text": "find me a flight from "
                            },
                            {
                                "entity": "city",
                                "slot_name": "origin",
                                "text": "Oslo"
                            },
                            {
                                "text": " to "
                            },
                            {
                                "entity": "city",
                                "slot_name": "destination",
                                "text": "Lima"
                            }
                        ]
                    },
                    {
                        "data": [
                            {
                                "text": "I need a flight leaving to "
                            },
                            {
                                "entity": "city",
                                "slot_name": "destination",
                                "text": "Berlin"
                            }
                        ]
                    }
                ]
            }

        Raises:
            IntentFormatError: When the YAML dict does not correspond to the
                :ref:`expected intent format <yaml_intent_format>`
        """
        if isinstance(yaml_dict, IOBase):
            yaml_dict = yaml.safe_load(yaml_dict)

        object_type = yaml_dict.get("type")
        if object_type and object_type != "intent":
            raise IntentFormatError("Wrong type: '%s'" % object_type)
        intent_name = yaml_dict.get("name")
        if not intent_name:
            raise IntentFormatError("Missing 'name' attribute")
        slot_mapping = dict()
        for slot in yaml_dict.get("slots", []):
            slot_mapping[slot["name"]] = slot["entity"]
        utterances = [IntentUtterance.parse(u.strip())
                      for u in yaml_dict["utterances"] if u.strip()]
        if not utterances:
            raise IntentFormatError(
                "Intent must contain at least one utterance")
        return cls(intent_name, utterances, slot_mapping)

    @classmethod
    def from_dic(cls, name, dic):
        slot_mapping = dict()
        utterances = list()
        for utterance in dic.get("utterances", []):
            data = utterance.get('data', [])
            for chunk in data:
                if chunk.get('slot_name'):
                    slot_mapping[chunk["slot_name"]] = chunk["entity"]
            utterances.append(IntentUtterance.from_dataset_dic(data))
        return cls(
            name,
            utterances,
            slot_mapping,
            version=dic.get('version'),
            id=dic.get('id')
        )

    @property
    def slots_sequences(self):
        tups = []
        for utterance in self.utterances:
            sequence = [
                (k, v)
                for k, v in Counter(
                    sc.slot_name
                    for sc in utterance.slot_chunks
                ).items()
            ]
            if sequence:
                tups.append(tuple(sequence))
        return set(tups)

    def _complete_slot_name_mapping(self):
        for utterance in self.utterances:
            for chunk in utterance.slot_chunks:
                if chunk.entity and chunk.slot_name not in self.slot_mapping:
                    self.slot_mapping[chunk.slot_name] = chunk.entity
        return self

    def _ensure_entity_names(self):
        for utterance in self.utterances:
            for chunk in utterance.slot_chunks:
                if chunk.entity:
                    continue
                chunk.entity = self.slot_mapping.get(
                    chunk.slot_name, chunk.slot_name)
        return self

    @property
    def json(self):
        """Intent data in json format"""
        return {
            UTTERANCES: [
                {DATA: [chunk.json for chunk in utterance.chunks]}
                for utterance in self.utterances
            ]
        }

    @property
    def entities_names(self):
        return set(chunk.entity for u in self.utterances
                   for chunk in u.chunks if isinstance(chunk, SlotChunk))

    def __str__(self):
        return "<%s name=%s nb_utterances=%s version=%s>" % (
            self.__class__.__name__,
            self.name,
            len(self.utterances),
            self.version,
        )


class IntentUtterance(utils.BaseObj):
    def __init__(self, chunks):
        self.chunks = chunks

    @property
    def text(self):
        return "".join((chunk.text for chunk in self.chunks))

    @property
    def tmpl(self):
        return ''.join(str(c) for c in self.chunks)

    @property
    def slot_chunks(self):
        return (chunk for chunk in self.chunks if isinstance(chunk, SlotChunk))

    @classmethod
    def parse(cls, string):
        """Parses an utterance

        Args:
            string (str): an utterance in the class:`.Utterance` format

        Examples:

            >>> from snips_nlu.dataset.intent import IntentUtterance
            >>> u = IntentUtterance.\
                parse("president of [country:default](France)")
            >>> u.text
            'president of France'
            >>> len(u.chunks)
            2
            >>> u.chunks[0].text
            'president of '
            >>> u.chunks[1].slot_name
            'country'
            >>> u.chunks[1].entity
            'default'
        """
        sm = SM(string)
        capture_text(sm)
        return cls(sm.chunks)

    @classmethod
    def from_dataset_dic(cls, dic_list):
        chunks = []
        for dic in dic_list:
            if dic.get('slot_name'):
                chunks.append(SlotChunk(
                    dic.get('slot_name'),
                    dic.get('entity'),
                    dic.get('text')
                ))
            else:
                chunks.append(TextChunk(dic.get('text')))
        return cls(chunks)

    def __str__(self):
        return "<%s '%s'>" % (
            self.__class__.__name__,
            self.tmpl
        )


class Chunk(metaclass=ABCMeta):
    def __init__(self, text):
        self.text = text

    @abstractmethod
    def json(self):
        pass


class SlotChunk(utils.BaseObj, Chunk):
    def __init__(self, slot_name, entity, text):
        super(SlotChunk, self).__init__(text)
        self.slot_name = slot_name
        self.entity = entity

    @property
    def json(self):
        return {
            TEXT: self.text,
            SLOT_NAME: self.slot_name,
            ENTITY: self.entity,
        }

    def __str__(self):
        return '[%s](%s)' % (self.slot_name, self.text)


class TextChunk(utils.BaseObj, Chunk):
    @property
    def json(self):
        return {
            TEXT: self.text
        }

    def __str__(self):
        return self.text


class SM(object):
    """State Machine for parsing"""

    def __init__(self, input):
        self.input = input
        self.chunks = []
        self.current = 0

    @property
    def end_of_input(self):
        return self.current >= len(self.input)

    def add_slot(self, name, entity=None):
        """Adds a named slot

        Args:
            name (str): slot name
            entity (str): entity name
        """
        chunk = SlotChunk(slot_name=name, entity=entity, text=None)
        self.chunks.append(chunk)

    def add_text(self, text):
        """Adds a simple text chunk using the current position"""
        chunk = TextChunk(text=text)
        self.chunks.append(chunk)

    def add_tagged(self, text):
        """Adds text to the last slot"""
        if not self.chunks:
            raise AssertionError("Cannot add tagged text because chunks list "
                                 "is empty")
        self.chunks[-1].text = text

    def find(self, s):
        return self.input.find(s, self.current)

    def move(self, pos):
        """Moves the cursor of the state to position after given

        Args:
            pos (int): position to place the cursor just after
        """
        self.current = pos + 1

    def peek(self):
        if self.end_of_input:
            return None
        return self[0]

    def read(self):
        c = self[0]
        self.current += 1
        return c

    def __getitem__(self, key):
        current = self.current
        if isinstance(key, int):
            return self.input[current + key]
        elif isinstance(key, slice):
            start = current + key.start if key.start else current
            return self.input[slice(start, key.stop, key.step)]
        else:
            raise TypeError("Bad key type: %s" % type(key))


def capture_text(state):
    next_pos = state.find('[')
    sub = state[:] if next_pos < 0 else state[:next_pos]
    if sub:
        state.add_text(sub)
    if next_pos >= 0:
        state.move(next_pos)
        capture_slot(state)


def capture_slot(state):
    next_colon_pos = state.find(':')
    next_square_bracket_pos = state.find(']')
    if next_square_bracket_pos < 0:
        raise IntentFormatError("Missing ending ']' in annotated utterance")
    if next_colon_pos < 0 or next_square_bracket_pos < next_colon_pos:
        slot_name = state[:next_square_bracket_pos]
        state.move(next_square_bracket_pos)
        state.add_slot(slot_name)
    else:
        slot_name = state[:next_colon_pos]
        state.move(next_colon_pos)
        entity = state[:next_square_bracket_pos]
        state.move(next_square_bracket_pos)
        state.add_slot(slot_name, entity)
    if state.peek() == '(':
        state.read()
        capture_tagged(state)
    else:
        capture_text(state)


def capture_tagged(state):
    next_pos = state.find(')')
    if next_pos < 1:
        raise IntentFormatError("Missing ending ')' in annotated utterance")
    else:
        tagged_text = state[:next_pos]
        state.add_tagged(tagged_text)
        state.move(next_pos)
        capture_text(state)



#!/usr/bin/env python3

"""
This module contains a Snips app that act as a middleware beween intent and
slots to remap.
"""

from configparser import ConfigParser
from copy import deepcopy
import pathlib
import json
import logging
import logging.handlers

import toml
import yaml

import paho.mqtt.client as mqtt


DEFAULT_TOML_PATH = '/etc/snips.toml'
ENCODING_FORMAT = "utf-8"
CONFIG_INI = "config.ini"
SLOTS = 'slots'
SLOT_NAME = 'slotName'
INTENT_NAME = "intentName"
HERMES_INTENT = "hermes/intent/%s"
TO = 'to'
INTENT = "intent"


log = logging.getLogger("snips-middleware")
log.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(filename)s:%(lineno)d %(levelname)-10s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
handler = logging.handlers.RotatingFileHandler('./middleware.log', encoding='utf-8')
handler.setLevel(logging.DEBUG)
handler.setFormatter(formatter)
log.addHandler(handler)
log.propagate = False


class SnipsConfigParser(ConfigParser):
    def to_dict(self):
        return {
            section: {
                option_name: option
                for option_name, option in self.items(section) if option.strip()
            }
            for section in self.sections()
        }


def read_configuration_file(configuration_file):
    try:
        with pathlib.Path(configuration_file).open(encoding=ENCODING_FORMAT) as f:
            conf_parser = SnipsConfigParser(delimiters=('='))
            conf_parser.optionxform = str
            conf_parser.readfp(f)
            return conf_parser.to_dict()
    except (IOError, ConfigParser.Error):
        return dict()


def load_route_table(path):
    with pathlib.Path(path).open('r') as fh:
        return yaml.load(fh, Loader=yaml.FullLoader)


class AssistantMiddleware(object):
    """ This app reroute intent and slots from your Snips assistant.

    The rerouting is based on you config.ini
    """
    def __init__(self, debug=False):
        """Initialize the app."""
        self.config = read_configuration_file(CONFIG_INI)['general']
        self.routing_table = load_route_table(self.config.get('routing_table'))
        self.debug = self.config.get('debug', False)
        try:
            mqtt_host_port = toml.load(DEFAULT_TOML_PATH)['snips-common']['mqtt']
            mqtt_host, mqtt_port = mqtt_host_port.split(':')
            mqtt_port = int(mqtt_port)
        except (KeyError, ValueError):
            # If the mqtt key doesn't exist or doesn't have the correct format,
            # use the default values.
            mqtt_host = self.config.get('mqtt_host', 'localhost')
            mqtt_port = int(self.config.get('mqtt_port', 1883))
        mqtt_timeout = int(self.config.get('mqtt_timeout', 5))

        self.client = mqtt.Client("snips-middleware")
        self.client.on_connect = self._subscribe_reroute
        if self.debug:
            log.info("debug mode activated")
            self.client.on_message = self.emit_rerouted_intent_debug

        self.client.connect(mqtt_host, mqtt_port, mqtt_timeout)
        self.client.loop_forever()

    def _subscribe_reroute(self, client, userdata, flags, rc):
        # remap intentA to intentB slot a to slot b
        if self.debug:
            self.client.subscribe(HERMES_INTENT % "#", 1)
        for intent_name, intent_dic in self.routing_table.items():
            new_intent_name = intent_dic.get(TO)
            log.info('hook re route intent : %s => %s', intent_name, new_intent_name)
            client.message_callback_add(
                HERMES_INTENT % intent_name, self.emit_rerouted_intent)
        log.info("started sucessfully")

    def emit_rerouted_intent_debug(self, client, userdata, message):
        log.debug("debug => captured intent")
        log.debug("topic: %s", str(message.topic))
        log.debug("payload: %s", str(message.payload))

    def emit_rerouted_intent(self, client, userdata, msg):
        payload = json.loads(msg.payload.decode(ENCODING_FORMAT))
        new_payload = deepcopy(payload)
        name = payload[INTENT][INTENT_NAME]

        intent_routing_info = self.routing_table[name]
        rerouted_intent_name = intent_routing_info.get(TO)
        rerouted_intent_name = self.route_intents[name]
        for slot_content in payload[SLOTS]:
            slot_name = slot_content[SLOT_NAME]
            if slot_name in intent_routing_info[SLOTS]:
                new_slot_name = intent_routing_info[SLOTS][slot_name]
                new_payload[SLOTS][new_slot_name] = slot_content
                del new_payload[SLOTS][slot_name]
        log.info("Intent %s detected with slots %s",
                 name,
                 [s[SLOT_NAME] for s in payload[SLOTS]])
        log.info("Rerouted to Intent %s detected with slots %s",
                 rerouted_intent_name,
                 [s[SLOT_NAME] for s in new_payload[SLOTS]])
        client.publish(
            'hermes/intent/%s' % rerouted_intent_name,
            json.dumps(new_payload)
        )


if __name__ == "__main__":
    AssistantMiddleware(debug=True)



# coding=utf-8

import json
import datetime

from . import Dataset
from .. import utils


class AssistantIntent(utils.BaseObj):

    def __init__(self, name, id, slots, version):
        self.name = name
        self.id = id
        self.slots = slots
        self.version = version

    @classmethod
    def load(cls, dic_info):
        return AssistantIntent(
            dic_info['name'],
            dic_info['id'],
            dic_info['slots'],
            dic_info['version'],
        )

    def __str__(self):
        return "<%s name=%s id=%s nb_slots=%s version=%s>" % (
            self.__class__.__name__,
            self.name,
            self.id,
            len(self.slots),
            self.version
        )


class Assistant(utils.BaseObj):

    """ Assistant Contract Specification """

    def __init__(self, name, language, versions, intents, created_at, original_path):
        self.name = name
        self.language = language
        self.versions = versions
        self._intents_list = intents
        self.created_at = created_at
        self._original_path = original_path
        self._dataset = None

    @property
    def intents(self):
        return {
            intent_spec.name: intent_spec
            for intent_spec in self._intents_list
        }

    @classmethod
    def load(cls, assistant_json):
        """ load specification from an assistant json file

        Args:
            assistant_json: pathlib.Path
        Return:
            Assistant
        """
        with assistant_json.open(encoding='utf8') as fh:
            assistant_spec = json.load(fh)

        return Assistant(
            name=assistant_spec['name'],
            created_at=assistant_spec['createdAt'],
            language=assistant_spec['language'],
            versions=assistant_spec.get('version', {}),
            intents=[
                AssistantIntent.load(_) for _ in assistant_spec.get('intents', [])
            ],
            original_path=assistant_json,
        )

    @property
    def dataset(self):
        if self._dataset:
            return self._dataset
        self._dataset = Dataset.from_json(
            self.language, self._original_path.parent / "dataset.json"
        )
        return self._dataset

    def __str__(self):
        return "<%s name='%s' lang='%s' nb_intents=%s versions=[%s] created_at=%s>" % (
            self.__class__.__name__,
            self.name,
            self.language,
            len(self.intents),
            ",".join("%s=%s" % (k, v) for k, v in self.versions.items()),
            str(
                datetime.datetime.strptime(
                    self.created_at, "%Y-%m-%dT%H:%M:%S.%fZ").date())
        )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


setup(
    name='snips-app-helpers',
    version='0.0.7',
    license='MIT license',
    description='Handful of cli and other tools to develop with confidence Snips services',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Dreamermind',
    author_email='sigmer66@gmail.com',
    url='https://github.com/dreamermind/snips-app-helpers',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    project_urls={
        'Documentation': 'https://snips-app-helpers.readthedocs.io/',
        'Changelog': 'https://snips-app-helpers.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/dreamermind/snips-app-helpers/issues',
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='!=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    install_requires=[
        'click',
        'colorama',
        'pyyaml',
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    entry_points={
        'console_scripts': [
            'snips-app = snips_app_helpers.cli:main',
        ]
    },
)



# coding=utf-8

from builtins import str
from io import IOBase

import yaml

from .constants import (
    AUTOMATICALLY_EXTENSIBLE, DATA, MATCHING_STRICTNESS, SYNONYMS,
    USE_SYNONYMS, VALUE)
from .exceptions import EntityFormatError
from ... import utils


class Entity(utils.BaseObj):
    """Entity data of a :class:`.Dataset`

    This class can represents both a custom or a builtin entity. When the
    entity is a builtin one, only the `name` attribute is relevant.

    Attributes:
        name (str): name of the entity
        utterances (list of :class:`.EntityUtterance`): entity utterances
            (only for custom entities)
        automatically_extensible (bool): whether or not the entity can be
            extended to values not present in the data (only for custom
            entities)
        use_synonyms (bool): whether or not to map entity values using
            synonyms (only for custom entities)
        matching_strictness (float): controls the matching strictness of the
            entity (only for custom entities). Must be between 0.0 and 1.0.
    """

    def __init__(self, name, utterances=None, automatically_extensible=True,
                 use_synonyms=True, matching_strictness=1.0, entity_type=None):
        if utterances is None:
            utterances = []
        self.name = name
        self.utterances = utterances
        self.automatically_extensible = automatically_extensible
        self.use_synonyms = use_synonyms
        self.matching_strictness = matching_strictness
        self.entity_type = entity_type

    @property
    def is_builtin(self):
        return self.name.startswith('snips/')

    @classmethod
    def from_yaml(cls, yaml_dict):
        """Build an :class:`.Entity` from its YAML definition object

        Args:
            yaml_dict (dict or :class:`.IOBase`): object containing the YAML
                definition of the entity. It can be either a stream, or the
                corresponding python dict.

        Examples:
            An entity can be defined with a YAML document following the schema
            illustrated in the example below:

            >>> import io
            >>> from snips_nlu.common.utils import json_string
            >>> entity_yaml = io.StringIO('''
            ... # City Entity
            ... ---
            ... type: entity
            ... name: city
            ... automatically_extensible: false # default value is true
            ... use_synonyms: false # default value is true
            ... matching_strictness: 0.8 # default value is 1.0
            ... values:
            ...   - london
            ...   - [new york, big apple]
            ...   - [paris, city of lights]''')
            >>> entity = Entity.from_yaml(entity_yaml)
            >>> print(json_string(entity.json, indent=4, sort_keys=True))
            {
                "automatically_extensible": false,
                "data": [
                    {
                        "synonyms": [],
                        "value": "london"
                    },
                    {
                        "synonyms": [
                            "big apple"
                        ],
                        "value": "new york"
                    },
                    {
                        "synonyms": [
                            "city of lights"
                        ],
                        "value": "paris"
                    }
                ],
                "matching_strictness": 0.8,
                "use_synonyms": false
            }

        Raises:
            EntityFormatError: When the YAML dict does not correspond to the
                :ref:`expected entity format <yaml_entity_format>`
        """
        if isinstance(yaml_dict, IOBase):
            yaml_dict = yaml.safe_load(yaml_dict)

        object_type = yaml_dict.get("type")
        if object_type and object_type != "entity":
            raise EntityFormatError("Wrong type: '%s'" % object_type)
        entity_name = yaml_dict.get("name")
        if not entity_name:
            raise EntityFormatError("Missing 'name' attribute")
        auto_extensible = yaml_dict.get(AUTOMATICALLY_EXTENSIBLE, True)
        use_synonyms = yaml_dict.get(USE_SYNONYMS, True)
        matching_strictness = yaml_dict.get("matching_strictness", 1.0)
        utterances = []
        for entity_value in yaml_dict.get("values", []):
            if isinstance(entity_value, list):
                utterance = EntityUtterance(entity_value[0], entity_value[1:])
            elif isinstance(entity_value, str):
                utterance = EntityUtterance(entity_value)
            else:
                raise EntityFormatError(
                    "YAML entity values must be either strings or lists, but "
                    "found: %s" % type(entity_value))
            utterances.append(utterance)

        return cls(name=entity_name,
                   utterances=utterances,
                   automatically_extensible=auto_extensible,
                   use_synonyms=use_synonyms,
                   matching_strictness=matching_strictness)

    @classmethod
    def from_dic(cls, name, dic):
        utterances = []
        for entity_value in dic.get("data", []):
            value = entity_value.get('value', "")
            synonyms = entity_value.get('synonyms', [])
            utterances.append(EntityUtterance(value, synonyms))
        return cls(name=name,
                   utterances=utterances,
                   automatically_extensible=dic.get(AUTOMATICALLY_EXTENSIBLE, True),
                   use_synonyms=dic.get(USE_SYNONYMS, True),
                   entity_type=dic.get('entity_type'),
                   matching_strictness=dic.get("matching_strictness", 1.0))

    @property
    def json(self):
        """Returns the entity in json format"""
        if self.is_builtin:
            return dict()
        return {
            AUTOMATICALLY_EXTENSIBLE: self.automatically_extensible,
            USE_SYNONYMS: self.use_synonyms,
            DATA: [u.json for u in self.utterances],
            MATCHING_STRICTNESS: self.matching_strictness
        }

    def __str__(self):
        return "<%s name='%s' use_synonyms=%s auto_ext=%s match_strict=%s>" % (
            self.__class__.__name__,
            self.name,
            self.use_synonyms,
            self.automatically_extensible,
            self.matching_strictness,
        )




class EntityUtterance(object):
    """Represents a value of a :class:`.CustomEntity` with potential synonyms

    Attributes:
        value (str): entity value
        synonyms (list of str): The values to remap to the utterance value
        """

    def __init__(self, value, synonyms=None):
        self.value = value
        if synonyms is None:
            synonyms = []
        self.synonyms = synonyms

    @property
    def variations(self):
        return [self.value] + self.synonyms

    @property
    def json(self):
        return {VALUE: self.value, SYNONYMS: self.synonyms}


def utf_8_encoder(f):
    for line in f:
        yield line.encode("utf-8")



# coding=utf-8

import datetime
from collections import defaultdict
from string import Formatter
from textwrap import TextWrapper

from .. import utils


class Report(object):

    def __init__(self, msgs):
        self.created_at = datetime.datetime.now()
        self.msgs = msgs

    @property
    def grouped_messages(self):
        grp = defaultdict(list)
        for msg in self.msgs:
            grp[msg.__class__].append(msg)
        return grp


# base messages

class Message(utils.BaseObj):
    STATIC_MSG = "NOT IMPLEMENTED"

    def __init__(self, **kwargs):
        self.__kwargs = sorted(kwargs.keys())
        for key, val in kwargs.items():
            setattr(self, key, val)

    @classmethod
    def _print_list_helper(cls, list_msg, message_list, helper_info=None):
        def kwargs_fn(msg):
            fieldnames = [
                fname
                for _, fname, _, _ in Formatter().parse(list_msg)
                if fname]
            return {
                fieldname: getattr(msg, fieldname)
                for fieldname in fieldnames
            }
        final_msg = "%s:\n" % cls.STATIC_MSG

        final_msg += '\t%s' % ("\n\t".join(
            "- " + list_msg.format(**kwargs_fn(msg)) for msg in message_list
        ))
        if helper_info:
            wrapper = TextWrapper()
            wrapper.initial_indent = "\t\t"
            wrapper.subsequent_indent = "\t\t"
            final_msg += '\n\tRemarks:\n%s' % wrapper.fill(helper_info)

        return final_msg

    @staticmethod
    def print_list(message_list):
        raise NotImplementedError("please add it in your message subclass")

    @property
    def _signature(self):
        return '-'.join(str(getattr(self, k)) for k in self.__kwargs)

    def __eq__(self, other):
        return self._signature == other._signature

    def __hash__(self):
        return hash(self._signature)

    def __str__(self):
        return "%s: %s" % (self.__class__.__name__, self.STATIC_MSG)


class System(Message):
    pass


class Info(Message):
    pass


class Warning(Message):
    pass


class Error(Message):
    pass

# dedicated Messages


class CorrectlyLinked(System):

    STATIC_MSG = 'Correctly linked intent to action code'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "@ {spec_filepath} applied to {action_dir} {intent_name}", message_list)


class DetectedSpec(Info):

    STATIC_MSG = 'Detected spec'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "@ {spec_filepath} applied to {action_dir}", message_list)


class NoSpec(Warning):

    STATIC_MSG = 'Missing spec for following actions'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper("{action_dir}", message_list)


class IntentNotInAssistant(Warning):

    STATIC_MSG = 'Action waiting intent not in assistant'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "{intent_name} from action: {action_name}",
            message_list,
            helper_info="This should not be a problem except that it consume "
            "resource with useless purpose"
        )

class GenericCoveredIntent(Warning):

    STATIC_MSG = 'Intents do not seem to be covered with slot pattern by action code'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "@ {spec_filepath} applied to {action_dir} intent {intent_name}",
            message_list,
            helper_info="This limit the possible analysis.",
        )


class NotCoveredIntent(Error):

    STATIC_MSG = 'Intents do not seem to be covered by any action code'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "{intent_name}",
            message_list,
            helper_info="This might be due to missing spec in some action codes\n"
            "else you should take it seriously as no response at all will be given"
            " by your assistant to final user.",
        )


class IncoherentAssistantDatasetIntent(Error):

    STATIC_MSG = 'Intents do not seem to be in dataset.json but is in assistant.json'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "{intent_name}",
            message_list
        )


class CoverageSlotSeq(Warning):

    STATIC_MSG = 'Intents do not seem to cover the slot sequence'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "@ {spec_filepath} intent {intent_name} {slots_sequences}",
            message_list
        )


class MissingSlot(Error):

    STATIC_MSG = 'Action code declared use of a slot NAME that does not exist in assistant'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "@ {spec_filepath} intent {intent_name} slot_name {slot_name} not in {assistant_slot_names}",
            message_list)


class InvalidSlotType(Error):

    STATIC_MSG = 'Action code declared use of a slot TYPE that is invalid in assistant'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "@ {spec_filepath} intent '{intent_name}' slot_name '{slot_name}' "
            "declared as type starting with '{slot_type}' but is '{expected_slot_type}' "
            "in assistant",
            message_list)


class IntentHookedMultipleTimes(Warning):

    STATIC_MSG = 'Some Intents seems to be hooked multiple times'

    @classmethod
    def print_list(cls, message_list):
        return cls._print_list_helper(
            "intent {intent_name} in actions: {action_names}",
            message_list,
            helper_info="While it might be legit do not forget that it means\n"
            "each time you trigger this intent n actions will be performed"
        )



# coding=utf-8
import io
from itertools import cycle
import json

import yaml

from .entity import Entity
from .intent import Intent
from .exceptions import DatasetFormatError
from ... import utils


def represent_none(self, _):
    return self.represent_scalar('tag:yaml.org,2002:null', '')


yaml.add_representer(type(None), represent_none)


class Dataset(utils.BaseObj):
    """Dataset used in the main NLU training API

    Consists of intents and entities data. This object can be built either from
    text files (:meth:`.Dataset.from_files`) or from YAML files
    (:meth:`.Dataset.from_yaml_files`).

    Attributes:
        language (str): language of the intents
        intents (list of :class:`.Intent`): intents data
        entities (list of :class:`.Entity`): entities data
    """

    def __init__(self, language, intents, entities):
        self.language = language
        self.intents = intents
        self.entities = entities
        self._add_missing_entities()
        self._ensure_entity_values()

    @classmethod
    def from_yaml_files(cls, language, filenames):
        """Creates a :class:`.Dataset` from a language and a list of YAML files
        or streams containing intents and entities data

        Each file need not correspond to a single entity nor intent. They can
        consist in several entities and intents merged together in a single
        file.

        Args:
            language (str): language of the dataset (ISO639-1)
            filenames (iterable): filenames or stream objects corresponding to
                intents and entities data.

        Example:

            A dataset can be defined with a YAML document following the schema
            illustrated in the example below:

            >>> import io
            >>> from snips_nlu.common.utils import json_string
            >>> dataset_yaml = io.StringIO('''
            ... # searchFlight Intent
            ... ---
            ... type: intent
            ... name: searchFlight
            ... slots:
            ...   - name: origin
            ...     entity: city
            ...   - name: destination
            ...     entity: city
            ...   - name: date
            ...     entity: snips/datetime
            ... utterances:
            ...   - find me a flight from [origin](Oslo) to [destination](Lima)
            ...   - I need a flight leaving to [destination](Berlin)
            ...
            ... # City Entity
            ... ---
            ... type: entity
            ... name: city
            ... values:
            ...   - london
            ...   - [paris, city of lights]''')
            >>> dataset = Dataset.from_yaml_files("en", [dataset_yaml])
            >>> print(json_string(dataset.json, indent=4, sort_keys=True))
            {
                "entities": {
                    "city": {
                        "automatically_extensible": true,
                        "data": [
                            {
                                "synonyms": [],
                                "value": "london"
                            },
                            {
                                "synonyms": [
                                    "city of lights"
                                ],
                                "value": "paris"
                            }
                        ],
                        "matching_strictness": 1.0,
                        "use_synonyms": true
                    }
                },
                "intents": {
                    "searchFlight": {
                        "utterances": [
                            {
                                "data": [
                                    {
                                        "text": "find me a flight from "
                                    },
                                    {
                                        "entity": "city",
                                        "slot_name": "origin",
                                        "text": "Oslo"
                                    },
                                    {
                                        "text": " to "
                                    },
                                    {
                                        "entity": "city",
                                        "slot_name": "destination",
                                        "text": "Lima"
                                    }
                                ]
                            },
                            {
                                "data": [
                                    {
                                        "text": "I need a flight leaving to "
                                    },
                                    {
                                        "entity": "city",
                                        "slot_name": "destination",
                                        "text": "Berlin"
                                    }
                                ]
                            }
                        ]
                    }
                },
                "language": "en"
            }

        Raises:
            DatasetFormatError: When one of the documents present in the YAML
                files has a wrong 'type' attribute, which is not 'entity' nor
                'intent'
            IntentFormatError: When the YAML document of an intent does not
                correspond to the
                :ref:`expected intent format <yaml_intent_format>`
            EntityFormatError: When the YAML document of an entity does not
                correspond to the
                :ref:`expected entity format <yaml_entity_format>`
        """
        entities = []
        intents = []
        for filename in filenames:
            if isinstance(filename, io.IOBase):
                intents_, entities_ = cls._load_dataset_parts(
                    filename, "stream object")
            else:
                with io.open(filename, encoding="utf8") as f:
                    intents_, entities_ = cls._load_dataset_parts(f, filename)
            intents += intents_
            entities += entities_
        return cls(language, intents, entities)

    @classmethod
    def from_json(cls, language, filepath):
        """ done the lazy way loading as json converting to yaml > """
        if not filepath.is_file():
            raise ValueError('Missing dataset file @ %s' % str(filepath))
        with filepath.open('r', encoding="utf8") as fh:
            dataset_dic = json.load(fh)
        entities = []
        intents = []
        for doc_type, docs in dataset_dic.items():
            if doc_type == "entities":
                for entity, item in docs.items():
                    entities.append(Entity.from_dic(entity, item))
            elif doc_type == "intents":
                for intent_name, item in docs.items():
                    intents.append(Intent.from_dic(intent_name, item))
        return cls(language, intents, entities)

    @classmethod
    def _load_dataset_parts(cls, stream, stream_description):
        intents = []
        entities = []
        for doc in yaml.safe_load_all(stream):
            doc_type = doc.get("type")
            if doc_type == "entity":
                entities.append(Entity.from_yaml(doc))
            elif doc_type == "intent":
                intents.append(Intent.from_yaml(doc))
            else:
                raise DatasetFormatError(
                    "Invalid 'type' value in YAML file '%s': '%s'"
                    % (stream_description, doc_type))
        return intents, entities

    def _add_missing_entities(self):
        entity_names = set(e.name for e in self.entities)

        # Add entities appearing only in the intents utterances
        for intent in self.intents:
            for entity_name in intent.entities_names:
                if entity_name not in entity_names:
                    entity_names.add(entity_name)
                    self.entities.append(Entity(name=entity_name))

    def _ensure_entity_values(self):
        entities_values = {entity.name: self._get_entity_values(entity)
                           for entity in self.entities}
        for intent in self.intents:
            for utterance in intent.utterances:
                for chunk in utterance.slot_chunks:
                    if chunk.text is not None:
                        continue
                    try:
                        chunk.text = next(entities_values[chunk.entity])
                    except StopIteration:
                        raise DatasetFormatError(
                            "At least one entity value must be provided for "
                            "entity '%s'" % chunk.entity)
        return self

    def _get_entity_values(self, entity):
        if entity.is_builtin:
            return []
        values = [v for utterance in entity.utterances
                  for v in utterance.variations]
        values_set = set(values)
        for intent in self.intents:
            for utterance in intent.utterances:
                for chunk in utterance.slot_chunks:
                    if not chunk.text or chunk.entity != entity.name:
                        continue
                    if chunk.text not in values_set:
                        values_set.add(chunk.text)
                        values.append(chunk.text)
        return cycle(values)

    @property
    def intent_per_name(self):
        return {
            intent.name.split(':')[-1]: intent
            for intent in self.intents
        }

    @property
    def json(self):
        """Dataset data in json format"""
        intents = {intent_data.intent_name: intent_data.json
                   for intent_data in self.intents}
        entities = {entity.name: entity.json for entity in self.entities}
        return dict(language=self.language, intents=intents, entities=entities)

    def __str__(self):
        return "<%s lang=%s nb_intents=%d nb_entities=%d>" % (
            self.__class__.__name__,
            self.language,
            len(self.intents),
            len(self.entities)
        )





class SnipsNLUError(Exception):
    """Base class for exceptions raised in the snips-nlu library"""


class DatasetFormatError(SnipsNLUError):
    """Raised when attempting to create a Snips NLU dataset using a wrong
    format"""


class EntityFormatError(DatasetFormatError):
    """Raised when attempting to create a Snips NLU entity using a wrong
    format"""


class IntentFormatError(DatasetFormatError):
    """Raised when attempting to create a Snips NLU intent using a wrong
    format"""
